from slurmise import job_database, slurm
from slurmise.config import SlurmiseConfiguration
from slurmise.fit.poly_fit import PolynomialFit

import numpy as np


class Slurmise:
    """
    API class for interacting with slurmise.
    """

    def __init__(self, toml_path=None):
        self.configuration = SlurmiseConfiguration(toml_path)

    def record(
        self,
        cmd: str,
        job_name: str | None = None,
        slurm_id: str | None = None,
        step_id: str | None = None,
    ):
        # Pull just the slurm ID
        metadata_json = slurm.parse_slurm_job_metadata(slurm_id=slurm_id, step_id=step_id)

        parsed_jd = self.configuration.parse_job_cmd(
            cmd=cmd,
            job_name=job_name,
            slurm_id=metadata_json["slurm_id"],
            step_id=metadata_json["step_id"],
        )

        parsed_jd.memory = metadata_json["max_rss"]
        parsed_jd.runtime = metadata_json["elapsed_seconds"]

        with job_database.JobDatabase.get_database(
            self.configuration.db_filename
        ) as database:
            database.record(parsed_jd)

    def raw_record(self, job_data):
        with job_database.JobDatabase.get_database(
            self.configuration.db_filename
        ) as database:
            database.record(job_data)

    def print(self):
        with job_database.JobDatabase.get_database(
            self.configuration.db_filename
        ) as database:
            database.print()

    def predict(self, cmd, job_name):
        query_jd = self.configuration.parse_job_cmd(cmd=cmd, job_name=job_name)
        query_jd = self.configuration.add_defaults(query_jd)
        query_model = PolynomialFit.load(
            query=query_jd, path=self.configuration.slurmise_base_dir
        )
        query_jd, query_warns = query_model.predict(query_jd)
        return query_jd, query_warns

    def update_model(self, cmd, job_name):
        query_jd = self.configuration.parse_job_cmd(cmd=cmd, job_name=job_name)
        with job_database.JobDatabase.get_database(
            self.configuration.db_filename
        ) as database:
            jobs = database.query(query_jd)

        self._update_model(query_jd, jobs)

    def _update_model(self, query_jd, jobs):
        model_path = self.configuration.slurmise_base_dir

        try:
            query_model = PolynomialFit.load(query=query_jd, path=model_path)
        except FileNotFoundError:
            query_model = PolynomialFit(query=query_jd, degree=2, path=model_path)

        random_state = np.random.RandomState(42)
        query_model.fit(jobs, random_state=random_state)

        query_model.save()

    def update_all_models(self):
        with job_database.JobDatabase.get_database(
            self.configuration.db_filename
        ) as database:
            for query_jd, jobs in database.iterate_database():
                self._update_model(query_jd, jobs)



import os

import h5py


def print_hdf5(
    h5py_obj, level=-1, print_full_name: bool = False, print_attrs: bool = True
) -> None:
    """Prints the name and shape of datasets in a H5py HDF5 file.

    Parameters
    ----------
    h5py_obj: [h5py.File, h5py.Group]
        the h5py.File or h5py.Group object
    level: int
        What level of the file tree you are in
    print_full_name
        If True, the full tree will be printed as the name, e.g. /group0/group1/group2/dataset: ...
        If False, only the current node will be printed, e.g. dataset:
    print_attrs
        If True: print all attributes in the file
    Returns
    -------
    None

    """

    def is_group(f):
        return type(f) == h5py._hl.group.Group

    def is_dataset(f):
        return type(f) == h5py._hl.dataset.Dataset

    def print_level(level, n_spaces=5) -> str:
        if level == -1:
            return ""
        prepend = "|" + " " * (n_spaces - 1)
        prepend *= level
        tree = "|" + "-" * (n_spaces - 2) + " "
        return prepend + tree

    for key in h5py_obj.keys():
        entry = h5py_obj[key]
        name = entry.name if print_full_name else os.path.basename(entry.name)
        if is_group(entry):
            print(f"{print_level(level)}{name}")
            print_hdf5(entry, level + 1, print_full_name=print_full_name)
        elif is_dataset(entry):
            shape = entry.shape
            dtype = entry.dtype
            print(f"{print_level(level)}{name}: {shape} {dtype} {entry[()]}")
    if level == -1:
        if print_attrs:
            print("attrs: ")
            for key, value in h5py_obj.attrs.items():
                print(f" {key}: {value}")



import os


def parse_slurmise_record_args(args: list[str]) -> dict:
    """
    Parse the arguments following the `slurmise record` command.

    Expects to receive a list of strings from the `ctx.args` object in the click context. This will have click recognized arguments removed.

    We only allow options to have one value, if multiple values are needed, they need to be quoted.

    For example, the following `slurmise record` command would be parsed:
    `cmd subcmd -o -k 2 -j -i 3 -m fast -q=5 file.json` into the following dictionary:

    :example:

                {
                    | 'cmd': ['cmd', 'subcmd', '-o', '-k', '2', '-j', '-i', '3', '-m', 'fast', '-q=5' ,'file.json'],
                    | 'positional': ['cmd', 'subcmd', 'file.json'],
                    | 'options': {'-k': '2', '-i': '3', '-m': 'fast', '-q': '5'},
                    | 'flags': {'-o': True, '-j': True}}

    """

    parsed_args = {
        "cmd": args,
        "positional": [],
        "options": {},
        "flags": {},
    }

    # Handle the flags and options
    prev_flag = None
    for i, arg in enumerate(args):
        if not arg.startswith("-"):
            if prev_flag:
                parsed_args["options"][prev_flag] = arg
                if prev_flag in parsed_args["flags"]:
                    del parsed_args["flags"][prev_flag]
                prev_flag = None
            else:
                parsed_args["positional"].append(arg)

        elif "=" in arg:
            flag, value = arg.split("=")  # NOTE assumes only 1 equals sign
            parsed_args["options"][flag] = value
        else:
            parsed_args["flags"][arg] = True
            prev_flag = arg

    return parsed_args


def process_slurmise_record_args(parsed_args: dict) -> dict:
    """
    Process the arguments following the `slurmise record` command.

    - Identify file paths and record metadata about the files
    - Convert argument values to appropriate types based on user configuration or type inference

    cmd
    `slurmise record --schema --name myjob --numerics n:3,q:17.4 --categories c:cat1,cat2 --flags --verbose`

    For example, the following parsed_args:
    {
        "cmd": ["sort", "-k", "2" "tests/sacct_output.json"],
        "positional": ["sort", "tests/sacct_output.json"],
        "options": {"-k": "2"},
        "flags": {},
    },

    would be processed into:
    [
        {
            "name": "sort",
            "arg_type": "positional",
            "type": "string",
            "value": "sort",
        },
        {
            "name": "tests/sacct_output.json",
            "arg_type": "positional",
            "type": "file",
            "size": 13398,
        },
        {"name": "-k", "arg_type": "option", "type": "string", "value": "-k"},
    ]
    """

    def process_value(value: str) -> dict:
        if os.path.exists(value):
            file_size = os.path.getsize(value)
            return {"type": "file", "size": file_size}

        # Try to convert to a numeric type (int, then float, then str)
        try:
            return {"type": "numeric", "value": int(value)}
        except ValueError:
            try:
                return {"type": "numeric", "value": float(value)}
            except ValueError:
                return {"type": "string", "value": value}

    # Loop through any positional arguments
    positional = [
        {"name": arg, "arg_type": "positional", **process_value(arg)}
        for arg in parsed_args["positional"]
    ]
    options = [
        {"name": arg, "arg_type": "option", **process_value(arg)}
        for arg in parsed_args["options"]
    ]
    flags = [
        {"name": arg, "arg_type": "flag", "value": True} for arg in parsed_args["flags"]
    ]

    return positional + options + flags



import pytest
from pathlib import Path
import shutil
from collections import namedtuple


TomlReturn = namedtuple('TomlReturn', ['toml', 'db'])


@pytest.fixture
def simple_toml(tmp_path):
    d = tmp_path
    p = d / "slurmise.toml"
    p.write_text(
        f"""
    [slurmise]
    base_dir = "{d/'slurmise_dir'}"

    [slurmise.job.nupack]
    job_spec = "monomer -T {{threads:numeric}} -C {{complexity:category}}"
    """
    )
    return TomlReturn(p, d / "slurmise_dir" / "slurmise.h5")


@pytest.fixture
def nupack_toml(tmp_path):
    d = tmp_path
    p = d / "slurmise.toml"
    p.write_text(
        f"""
    [slurmise]
    base_dir = "{d/'slurmise_dir'}"
    db_filename = "nupack2.h5"

    [slurmise.job.nupack]
    job_spec = "monomer -c {{cpus:numeric}} -S {{sequences:numeric}}"
    """
    )

    db_path = d / "slurmise_dir" / "nupack2.h5"
    Path.mkdir(db_path.parent, exist_ok=True, parents=True)
    shutil.copyfile(
        "./tests/nupack2.h5",
        db_path,
    )

    return TomlReturn(p, db_path)





from dataclasses import astuple, dataclass, field

import h5py
import numpy as np


def array_safe_eq(a, b) -> bool:
    """
    Check if a and b are equal, even if they are numpy arrays.
    When a and be are dictionaries call recursively for all key, value pairs.
    """

    if a is b:
        return True
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and (a == b).all()
    if isinstance(a, dict) and isinstance(b, dict):
        return a.keys() == b.keys() and all(
            array_safe_eq(a[key], b[key]) for key in a.keys()
        )
    try:
        return a == b
    except TypeError:
        return NotImplemented


def dc_eq(dc1, dc2) -> bool:
    """
    Checks if two dataclasses which hold numpy arrays are equal
    """

    if dc1 is dc2:
        return True
    if dc1.__class__ is not dc2.__class__:
        return NotImplemented  # better than False
    t1 = astuple(dc1)
    t2 = astuple(dc2)
    return all(array_safe_eq(a1, a2) for a1, a2 in zip(t1, t2))


@dataclass(eq=False)
class JobData:
    """
    Jobdata class holds the information of a unique slurm job.

    :arguments:

        :job_name: The unique command name to execute under slurm.
        :slurm_id: The slurm job id assigned by the sceduler for a job run.
        :categorical: the CLI parameters of this job. This has parameter that affect the performance of the job and are fit seperately.
        :numerical: These are parameters that are used as the free variables for fits, such input size, number of iterations etc.
        :memory: The maximum amount of memory in MBs this job used.
        :runtime: The time this job needed to complete in minutes.
    """

    job_name: str
    slurm_id: str | None = None
    categorical: dict = field(default_factory=dict)
    numerical: dict = field(default_factory=dict)
    memory: int | None = None  # in MBs
    runtime: int | None = None  # in minutes
    cmd: str | None = None  # TODO: NOT STORED OR RETURNED

    @staticmethod
    def from_dataset(
        job_name: str, slurm_id: str, dataset: h5py.Dataset, categorical: dict
    ) -> "JobData":
        """
        This method creates a JobData object from a HDF5 dataset that describes a job.
        :arguments:

            :job_name: The unique command name to execute under slurm.
            :slurm_id: The slurm job id assigned by the sceduler for a job run.
            :dataset: The HDF5 dataset used to populate numerical, memory and runtime information of the job.
        """

        runtime = dataset.get("runtime", None)
        if runtime is not None:
            runtime = runtime[()]
        memory = dataset.get("memory", None)
        if memory is not None:
            memory = memory[()]
        numerical = {
            key: value[()]
            for key, value in dataset.items()
            if key not in ("runtime", "memory")
        }
        categorical = dict(**categorical)

        return JobData(
            job_name=job_name,
            slurm_id=slurm_id,
            numerical=numerical,
            categorical=categorical,
            memory=memory,
            runtime=runtime,
        )

    def __eq__(self, other):
        return dc_eq(self, other)



import pytest
from slurmise.config import SlurmiseConfiguration
from slurmise.job_parse import file_parsers


@pytest.fixture
def basic_toml(tmpdir):
    d = tmpdir.mkdir("slurmise_dir")
    f = d.join("basic.toml")

    toml_str = """
    [slurmise]
    base_dir = "slurmise_dir"
    default_mem = 2000
    default_time = 70

    [slurmise.job.nupack]
    job_spec = "monomer -T {threads:numeric} -C {complexity:category}"
    default_mem = 3000
    default_time = 80

    [slurmise.job.with_ignore]
    job_prefix = "nothing"
    job_spec = "-T {threads:numeric} -C {complexity:category} -i {ignore}"

    # builtins will include file_size and file_lines
    # specify custom options here
    [slurmise.file_parsers.get_epochs]
    return_type = "numerical"
    awk_script = "'/^epochs:/ {print $2}'"

    [slurmise.file_parsers.fasta_lengths]
    return_type = "numerical"
    awk_script = "/a/path/to/file"
    script_is_file = true

    # categorical default return type
    [slurmise.file_parsers.script_string]
    awk_script = "/^>/"
    script_is_file = false

    # this is ignored in parsing as the argument doesn't match an awk parser
    [slurmise.file_parsers.unknown_type]
    no_awk_script = "/^>/"
    script_is_file = false
    """

    f.write(toml_str)
    return f


@pytest.fixture
def basic_toml_no_default(tmpdir):
    d = tmpdir.mkdir("slurmise_dir")
    f = d.join("basic.toml")

    toml_str = """
    [slurmise]
    base_dir = "slurmise_dir"

    [slurmise.job.nupack]
    job_spec = "monomer -T {threads:numeric} -C {complexity:category}"
    """

    f.write(toml_str)
    return f

def test_init_SlurmiseConfiguration(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    assert config.slurmise_base_dir == "slurmise_dir"
    assert len(config.jobs) == 2
    assert config.jobs['with_ignore']['job_prefix'] == "nothing"


def test_parse_job_cmd(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    job_data = config.parse_job_cmd("monomer -T 1 -C simple", "nupack", "1234")

    assert job_data.job_name == "nupack"
    assert job_data.slurm_id == "1234"
    assert job_data.categorical == {"complexity": "simple"}
    assert job_data.numerical == {"threads": 1}


def test_parse_job_cmd_with_ignore(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    job_data = config.parse_job_cmd("-T 1 -C simple -i can't see me", "with_ignore", "1234")

    assert job_data.job_name == "with_ignore"
    assert job_data.slurm_id == "1234"
    assert job_data.categorical == {"complexity": "simple"}
    assert job_data.numerical == {"threads": 1}

def test_parse_job_cmd_invalid(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    with pytest.raises(ValueError, match="Job spec monomer -T {threads:numeric} -C {complexity:category} does not match command dimer -T 1 -C simple."):
        config.parse_job_cmd("dimer -T 1 -C simple", "nupack", "1234")

def test_parse_job_cmd_name_mismatch(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    with pytest.raises(ValueError, match="Job oldpack not found in configuration."):
        config.parse_job_cmd("monomer -T 1 -C simple", "oldpack", "1234")

def test_parse_job_cmd_invalid_numeric(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    with pytest.raises(ValueError, match="Job spec monomer -T {threads:numeric} -C {complexity:category} does not match command monomer -T 1A -C simple."):
        config.parse_job_cmd("monomer -T 1A -C simple", "nupack", "1234")

def test_awk_parsers(basic_toml):
    config = SlurmiseConfiguration(basic_toml)

    assert config.file_parsers == {
        'file_size': file_parsers.FileSizeParser(),
        'file_lines': file_parsers.FileLinesParser(),
        'file_basename': file_parsers.FileBasename(),
        'file_md5': file_parsers.FileMD5(),
        'get_epochs': file_parsers.AwkParser('get_epochs', 'numerical', "'/^epochs:/ {print $2}'", False),
        'fasta_lengths': file_parsers.AwkParser('fasta_lengths', 'numerical', "/a/path/to/file", True),
        'script_string': file_parsers.AwkParser('script_string', 'categorical', "/^>/", False),
    }

def test_parse_job_cmd_inference(basic_toml):
    config = SlurmiseConfiguration(basic_toml)
    with pytest.raises(ValueError, match="Unable to match job name to 'sort infile'"):
        config.parse_job_cmd('sort infile')

    match_prefix = config.parse_job_cmd('nothing -T 3 -C high -i something')
    assert match_prefix.job_name == 'with_ignore'

    match_name = config.parse_job_cmd('nupack monomer -T 3 -C high')
    assert match_name.job_name == 'nupack'

def test_default_resources_slurmise_base(basic_toml):
    '''Test the default can be set at the slurmise level for all jobs without additional defaults.'''
    config = SlurmiseConfiguration(basic_toml)
    job_data = config.parse_job_cmd('nothing -T 3 -C high -i something')
    config.add_defaults(job_data)

    assert job_data.memory == 2000
    assert job_data.runtime == 70

    # nupack job has defaults overwritten
    job_data = config.parse_job_cmd('nupack monomer -T 3 -C high')
    config.add_defaults(job_data)

    assert job_data.memory == 3000
    assert job_data.runtime == 80


def test_default_resources_no_setting(basic_toml_no_default):
    '''Test the default can be set at the slurmise level for all jobs without additional defaults.'''
    config = SlurmiseConfiguration(basic_toml_no_default)
    job_data = config.parse_job_cmd('nupack monomer -T 3 -C high')
    config.add_defaults(job_data)

    assert job_data.memory == 1000
    assert job_data.runtime == 60



from pathlib import Path

import numpy as np
import pytest

from slurmise.fit.poly_fit import PolynomialFit
from slurmise.fit.resource_fit import ResourceFit
from slurmise.job_data import JobData
from slurmise.job_database import JobDatabase


@pytest.fixture(autouse=True)
def monkey_patch_basepath(tmp_path, monkeypatch):
    """
    Monkey patch the BASEMODEL_PATH to the tmp_path, don't want to write to the actual
    path during testing (probably is user's home directory or something)
    """
    monkeypatch.setattr("slurmise.fit.resource_fit.BASEMODELPATH", tmp_path)
    yield
    monkeypatch.undo()


@pytest.mark.parametrize("specify_path", [True, False])
def test_model_path_creation(tmp_path, specify_path):
    """Test that both default and user-specified model paths are created correctly"""
    query = JobData(job_name="nupack")

    fit = ResourceFit(query=query, path=tmp_path if specify_path else None)

    # Check if the auto-generated path contains the class anem and the hash of the query
    if not specify_path:
        assert str(
            Path(fit.__class__.__name__) / Path(fit._get_model_info_hash(query))
        ) in str(fit.path)
    else:
        assert tmp_path == fit.path


@pytest.fixture(scope="module")
def nupack_data():
    query = JobData(job_name="nupack")

    with JobDatabase.get_database("tests/nupack2.h5") as db:
        # Get the job data
        jobs = db.query(job_data=query)

    # Drop jobs with 0 runtime or memory
    jobs = [job for job in jobs if job.runtime > 0 and job.memory > 0]

    # Only take jobs where sequences is len = 10
    # jobs = [job for job in jobs if job.numerical['sequences'].shape[0] == 10]

    return query, jobs


@pytest.mark.parametrize(
    "model, kwargs, expected_metrics",
    [
        (
            PolynomialFit,
            {"degree": 2},
            {
                "runtime": {"mpe": 28.5158571, "mse": 27.5574959},
                "memory": {"mpe": 18.0107243, "mse": 121437.5881779},
            },
        )
    ],
)
def test_fit_and_predict(nupack_data, model, kwargs, expected_metrics):
    """Test the fit classes on the nupack data"""

    query, jobs = nupack_data

    poly_fit = model(query=query, **kwargs)

    random_state = np.random.RandomState(42)
    poly_fit.fit(jobs, random_state=random_state)

    # Predict the runtime and memory of a job
    job = jobs[0]

    predicted_job, _ = poly_fit.predict(job)

    assert poly_fit.last_fit_dsize == int(len(jobs) * 0.8)

    # Save the model
    poly_fit.save()

    # Load the model
    poly_fit_loaded = PolynomialFit.load(query=query)
    assert poly_fit_loaded.last_fit_dsize == int(len(jobs) * 0.8)
    predicted_job2, _ = poly_fit_loaded.predict(job)
    assert predicted_job.runtime == predicted_job2.runtime
    assert predicted_job.memory == predicted_job2.memory

    for key in poly_fit.model_metrics.keys():
        assert key in expected_metrics
        for metric in poly_fit.model_metrics[key].keys():
            assert metric in expected_metrics[key]
            np.testing.assert_allclose(
                poly_fit.model_metrics[key][metric],
                expected_metrics[key][metric],
                rtol=1e-6,
            )



import json
import os
import subprocess


def parse_slurm_job_metadata(slurm_id: str | None = None, step_name: str | None = None) -> dict:
    """
    Return a dictionary of metadata for the current SLURM job.
    Parameters:
        slurm_id (str | None): The SLURM job ID. If None, the function will attempt to retrieve
            the job ID from the SLURM_JOBID environment variable.
        step_id (str | None): The SLURM step ID. If None, the function defaults to the last step
            of the job. If provided, it specifies which step's metadata to return.
    Returns:
        dict: A dictionary containing metadata for the specified SLURM job and step.
    """
    
    sacct_json = get_slurm_job_sacct(slurm_id)
    
    try:
        job_id = sacct_json["jobs"][0]["job_id"]
        job_name = sacct_json["jobs"][0]["name"]
        state = sacct_json["jobs"][0]["state"]["current"][0]
        partition = sacct_json["jobs"][0]["partition"]
        CPUs = sacct_json["jobs"][0]["required"]["CPUs"]
        memory_per_cpu = sacct_json["jobs"][0]["required"]["memory_per_cpu"]
        memory_per_node = sacct_json["jobs"][0]["required"]["memory_per_node"]
        max_rss = 0
        steps = {}
        jobstep_ids = []
        for step in sacct_json["jobs"][0]["steps"]:
            steps[step["step"]["id"]] = step
            jobstep_ids.append(step["step"]["id"])


        if step_name is None:
            step_id = jobstep_ids[-1]
            step_name = step_id.split(".")[-1]
        else:
            step_id = f"{job_id}.{step_name}"

        # In addition, the max requested memory is updated as slurm steps are completed.
        elapsed_seconds = int(steps[step_id]["time"]["elapsed"])
        for item in steps[step_id]["tres"]["requested"]["max"]:
            if item["type"] == "mem":
                max_rss = max(max_rss, item["count"] // (2**20))  # convert to MB
    except Exception as e:
        raise ValueError(
            f"Could not parse json from sacct cmd:\n\n {sacct_json}"
        ) from e

    metadata = {
        "slurm_id": job_id,
        "step_id": step_name,
        "job_name": job_name,
        "state": state,
        "partition": partition,
        "elapsed_seconds": elapsed_seconds,
        "CPUs": CPUs,
        "memory_per_cpu": memory_per_cpu,
        "memory_per_node": memory_per_node,
        "max_rss": max_rss,
    }
    return metadata


def get_slurm_job_sacct(slurm_id: str | None = None) -> dict:
    """Return the JSON output of the sacct command for the current SLURM job."""
    if slurm_id is None:
        if "SLURM_JOBID" not in os.environ:
            raise ValueError("Not running in a SLURM job")
        slurm_id = os.environ["SLURM_JOBID"]

    try:
        json_encoded_str = subprocess.check_output(["sacct", "-j", slurm_id, "--json"])
    except subprocess.CalledProcessError as e:
        raise ValueError(f"Error running sacct cmd: {e}") from e

    return json.loads(json_encoded_str.decode())



import pytest

from slurmise import parse_args


# @pytest.mark.parametrize(
#     "args",
#     [
#         "record cmd subcmd -o -k 2 -j -i 3 -m fast -q=5",
#     ],
# )
# def test_record_subcommand(args, monkeypatch):
#     """Use click test runner to ensure record subcommand exits with errorcode 0."""

#     def mock_get_slurm_job_sacct():
#         with open("tests/sacct_output.json") as f:
#             return json.load(f)

#     def mock_get_slurm_job_sstat():
#         return {}

#     monkeypatch.setattr(slurm, "get_slurm_job_sacct", mock_get_slurm_job_sacct)
#     monkeypatch.setattr(slurm, "get_slurm_job_sstat", mock_get_slurm_job_sstat)

#     runner = CliRunner()
#     result = runner.invoke(main, args.split())

#     assert result.exit_code == 0


@pytest.mark.parametrize(
    "args,exp",
    [
        (
            ["sleep", "2"],
            {
                "cmd": ["sleep", "2"],
                "positional": ["sleep", "2"],
                "options": {},
                "flags": {},
            },
        ),
        (
            ["sort", "-k", "2"],
            {
                "cmd": ["sort", "-k", "2"],
                "positional": ["sort"],
                "options": {"-k": "2"},
                "flags": {},
            },
        ),
        (
            ["time", "-l", "-p", "2", "--dog"],
            {
                "cmd": ["time", "-l", "-p", "2", "--dog"],
                "positional": ["time"],
                "options": {"-p": "2"},
                "flags": {"-l": True, "--dog": True},
            },
        ),
        (
            ["samtools", "view", "chr:1-100", "-b", "-o", "output.bam"],
            {
                "cmd": ["samtools", "view", "chr:1-100", "-b", "-o", "output.bam"],
                "positional": ["samtools", "view", "chr:1-100"],
                "options": {"-o": "output.bam"},
                "flags": {"-b": True},
            },
        ),
        (
            ["grep", "--color=auto", "-i", "hello world"],
            {
                "cmd": ["grep", "--color=auto", "-i", "hello world"],
                "positional": ["grep"],
                "options": {"--color": "auto", "-i": "hello world"},
                "flags": {},
            },
        ),
        (
            ["jq", "-l", ".[]", "file.json"],
            {
                "cmd": ["jq", "-l", ".[]", "file.json"],
                "positional": ["jq", "file.json"],
                "options": {"-l": ".[]"},
                "flags": {},
            },
        ),
    ],
)
def test_parse_record_args(args, exp):
    assert parse_args.parse_slurmise_record_args(args) == exp


@pytest.mark.parametrize(
    "parsed_args,exp",
    [
        (
            {
                "cmd": ["sort", "-k", "2" "tests/sacct_output.json"],
                "positional": ["sort", "tests/sacct_output.json"],
                "options": {"-k": "2"},
                "flags": {},
            },
            [
                {
                    "name": "sort",
                    "arg_type": "positional",
                    "type": "string",
                    "value": "sort",
                },
                {
                    "name": "tests/sacct_output.json",
                    "arg_type": "positional",
                    "type": "file",
                    "size": 13398,
                },
                {"name": "-k", "arg_type": "option", "type": "string", "value": "-k"},
            ],
        ),
    ],
)
def test_process_record_args(parsed_args, exp):
    processed_args = parse_args.process_slurmise_record_args(parsed_args)
    print(processed_args)
    assert processed_args == exp



import hashlib
import json
import pathlib
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Optional

import numpy as np

from ..job_data import JobData

BASEMODELPATH = pathlib.Path.home() / ".slurmise/models/"


@dataclass(kw_only=True)
class ResourceFit:
    query: JobData
    last_fit_dsize: int = 0
    fit_timestamp: datetime = field(default_factory=datetime.now)
    model_metrics: dict = field(default_factory=dict)
    path: Optional[pathlib.Path] = None

    def __post_init__(self):
        if isinstance(self.path, str):
            self.path = pathlib.Path(self.path)
        elif isinstance(self.path, pathlib.Path):
            pass
        elif self.path is None:
            self.path = self._make_model_path(query=self.query)
        else:
            raise ValueError("path must be a string or pathlib.Path object")

    @classmethod
    def _get_model_info_hash(cls, query: JobData) -> str:
        """
        This method generates a hash of the model's query information.
        """
        hash_info = {
            "class": cls.__name__,
            "job_name": query.job_name,
            **query.categorical,
        }
        hash_info_tuple = tuple(hash_info.items())

        # Get and MD5 hash of information
        hash_val = hashlib.md5(str(hash_info_tuple).encode("utf-8")).hexdigest()

        return hash_val

    @classmethod
    def _make_model_path(cls, query) -> pathlib.Path:
        """
        This method returns the path to the model's directory.

        The model's path is a function of the model's type and the hash of its query.
        """
        hash_val = cls._get_model_info_hash(query)
        return pathlib.Path(BASEMODELPATH) / cls.__name__ / hash_val

    def save(self, model_params: dict = {}):
        """This method saves the basic information of the model, such as its query,
        when it was last fit, the dataset size of the latest fit, and the type of
        the model.
        """
        self.path.mkdir(parents=True, exist_ok=True)
        with open(str(self.path / "fits.json"), "w") as save_file:
            # This converts the dataclass to a dictionary. If it is called from a subclass,
            # the subclass's attributes will be included in the dictionary.
            info = asdict(self)

            # Convert path to string
            info["path"] = str(info["path"])

            # Convert datetime to string
            info["fit_timestamp"] = info["fit_timestamp"].isoformat()

            info.update(model_params)
            json.dump(info, save_file)

    @classmethod
    def load(
        cls, query: JobData | None = None, path: str | None = None
    ) -> "ResourceFit":
        """
        This method loads a model from a file. The model is loaded from the path
        provided, or from the path generated from the query.

        :param query: The query used to generate the model
        :type query: JobData
        :param path: The path to the model
        :type path: str
        :return: The model
        :rtype: ResourceFit

        """
        match [path, query]:
            case (None, None):
                raise ValueError("Either query or path must be provided")
            case (None, _):
                path = cls._make_model_path(query)
            case (str(path), _):
                path = pathlib.Path(path)

        with open(str(path / "fits.json")) as load_file:
            info = json.load(load_file)

        # Convert datetime from isoformat string to datetime object
        info["fit_timestamp"] = datetime.fromisoformat(info["fit_timestamp"])

        # Generates an instance of the class from the dictionary. When this is called by
        # a ResourceFit subclass, it includes all attributes of the subclass(es) and the
        # ResourceFit class.

        return cls(**info)

    def predict(self, job: JobData) -> tuple[JobData, list[str]]:
        raise NotImplementedError

    def fit(
        self, jobs: list[JobData], random_state: np.random.RandomState | None, **kwargs
    ):
        raise NotImplementedError



from slurmise.api import Slurmise

import multiprocessing
import time
import pytest
from unittest import mock


def slurmise_record(toml, process_id, error_queue):
    def mock_metadata(kwargs):
        result = {
            "slurm_id": kwargs["slurm_id"],
            "job_name": "nupack",
            "state": "COMPLETED",
            "partition": "",
            "elapsed_seconds": 97201,
            "CPUs": 1,
            "memory_per_cpu": 0,
            "memory_per_node": 0,
            "max_rss": 232,
            "step_id": "external",
        }
        return result

    try:
        time.sleep(process_id * 0.1)
        with mock.patch(
            "slurmise.slurm.parse_slurm_job_metadata",
            side_effect=lambda *args, **kwargs: mock_metadata(kwargs),
        ):
            slurmise = Slurmise(toml)
            time.sleep(process_id * 0.1)
            for i in range(10):
                slurmise.record(
                    "nupack monomer -T 2 -C simple", slurm_id=process_id * 100 + i
                )
                time.sleep(process_id * 0.1)
    except Exception as e:
        error_queue.put(f"PID {process_id}: {e}")


def test_multiple_slurmise_instances(simple_toml):
    processes = []
    error_queue = multiprocessing.Queue()
    for i in range(10):
        p = multiprocessing.Process(
            target=slurmise_record, args=(simple_toml.toml, i, error_queue)
        )
        processes.append(p)
        p.start()

    [p.join() for p in processes]

    if not error_queue.empty():
        while not error_queue.empty():
            print(error_queue.get())
        pytest.fail("Child prcess had error")



from dataclasses import InitVar, dataclass

import joblib
import numpy as np
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

# Generate a polynomial fit for the runtime data using sklearn
from sklearn.preprocessing import OneHotEncoder, PolynomialFeatures, StandardScaler

from slurmise.job_data import JobData
from slurmise.utils import jobs_to_pandas
from slurmise.fit.resource_fit import ResourceFit


@dataclass(kw_only=True)
class PolynomialFit(ResourceFit):
    degree: int
    runtime_model: InitVar[Pipeline | None] = None
    memory_model: InitVar[Pipeline | None] = None

    def __post_init__(self, runtime_model, memory_model):
        self.runtime_model = runtime_model
        self.memory_model = memory_model

        super().__post_init__()

    def save(self):
        super().save()
        if self.runtime_model is not None:
            modelpath = self.path / "runtime_model.pkl"
            joblib.dump(self.runtime_model, str(modelpath))
        if self.memory_model is not None:
            modelpath = self.path / "memory_model.pkl"
            joblib.dump(self.memory_model, str(modelpath))

    @classmethod
    def load(cls, query: JobData | None = None, path: str | None = None):
        fit_obj = super().load(query=query, path=path)

        fit_obj.runtime_model = joblib.load(str(fit_obj.path / "runtime_model.pkl"))
        fit_obj.memory_model = joblib.load(str(fit_obj.path / "memory_model.pkl"))

        return fit_obj

    def _make_model(self, categorical_features, numerical_features):
        categorical_transformer = Pipeline(
            steps=[
                ("encoder", OneHotEncoder(handle_unknown="infrequent_if_exist")),
            ]
        )
        numeric_transformer = Pipeline(
            steps=[
                ("imputer", SimpleImputer(strategy=np.max)),
                ("scaler", StandardScaler()),
            ]
        )

        preprocessor = ColumnTransformer(
            transformers=[
                ("num", numeric_transformer, numerical_features),
                ("cat", categorical_transformer, categorical_features),
            ]
        )

        # We are doing polynomial regression, so we need to add polynomial features
        poly = PolynomialFeatures(degree=self.degree, include_bias=False)
        model = LinearRegression()

        pipeline = Pipeline(
            [("preprocessor", preprocessor), ("poly", poly), ("model", model)]
        )

        return pipeline

    def fit(
        self, jobs: list[JobData], random_state: np.random.RandomState | None, **kwargs
    ):
        X, categorical_features, numerical_features = jobs_to_pandas(jobs)

        Y = X[["runtime", "memory"]]

        # Drop the runtime and memory columns
        X = X.drop(columns=["runtime", "memory"])

        # Split test and train data
        X_train, X_test, y_train, y_test = train_test_split(
            X, Y, test_size=0.2, random_state=random_state
        )

        self.last_fit_dsize = len(X_train)

        self.runtime_model = self._make_model(categorical_features, numerical_features)
        self.memory_model = self._make_model(categorical_features, numerical_features)

        self.runtime_model.fit(X_train, y_train["runtime"])
        self.memory_model.fit(X_train, y_train["memory"])

        # Evaluate the model on test
        Y_pred_runtime = self.runtime_model.predict(X_test)
        Y_pred_memory = self.memory_model.predict(X_test)

        def mean_percent_error(y_true, y_pred):
            return np.mean(np.abs((y_true - y_pred) / y_true)) * 100

        self.model_metrics = {
            "runtime": {
                "mpe": mean_percent_error(y_test["runtime"], Y_pred_runtime),
                "mse": mean_squared_error(y_test["runtime"], Y_pred_runtime),
            },
            "memory": {
                "mpe": mean_percent_error(y_test["memory"], Y_pred_memory),
                "mse": mean_squared_error(y_test["memory"], Y_pred_memory),
            },
        }
        # TODO: Warning if model metrics are larger than a threshold.

    def predict(self, job: JobData) -> tuple[JobData, list[str]]:
        # TODO: check if it can be abstracted.

        if self.last_fit_dsize < 10:

            return (
                job,
                "Not enough fitting data points in the fits. Returning default values.",
            )

        X, _, _ = jobs_to_pandas([job])
        warnmsg = []
        if self.model_metrics["runtime"]["mpe"] < 10:
            warnmsg += [
                f"Runtime prediction for job {job.job_name} is not within 10% of actual value.",
                "Returing default runtime value.",
            ]
        else:
            predicted_runtime = self.runtime_model.predict(X)[0]
            if predicted_runtime > 0 and predicted_runtime < 100 * job.runtime:
                job.runtime = predicted_runtime
            else:
                warnmsg += [
                    f"Predicted runtime for job {job.job_name} is either negative or more than 100 times larger than default.",
                    "Returing default runtime value.",
                ]

        if self.model_metrics["memory"]["mpe"] < 10:
            warnmsg += [
                f"Memory prediction for job {job.job_name} is not within 10% of actual value.",
                "Returing default memory value.",
            ]
        else:
            predicted_memory = self.memory_model.predict(X)[0]
            if predicted_memory > 0 and predicted_memory < 100 * job.memory:
                job.memory = predicted_memory
            else:
                warnmsg += [
                    f"Predicted memory for job {job.job_name} is either negative or more than 100 times larger than default.",
                    "Returing default memory value.",
                ]

        return job, warnmsg



import json
import sys

import click

from slurmise import job_data
from slurmise.api import Slurmise


@click.group()
@click.option(
    "--toml",
    "-t",
    type=click.Path(exists=True),
    required=False,
    help="Path to the slurmise configuration file",
)
@click.pass_context
def main(ctx, toml):
    if toml is None:
        click.echo("Slurmise requires a toml file", err=True)
        click.echo("See readme for more information", err=True)
        sys.exit(1)
    ctx.ensure_object(dict)
    ctx.obj["slurmise"] = Slurmise(toml)


@main.command()
@click.argument("cmd", nargs=1)
@click.option("--job-name", type=str, help="Name of the job")
@click.option("--slurm-id", type=str, help="SLURM id of job")
@click.option("--step-id", type=str, help="SLURM step id")
@click.pass_context
def record(ctx, cmd, job_name, slurm_id, step_id):
    """Command to record a job.
    For example: `slurmise record "-o 2 -i 3 -m fast"`
    """

    ctx.obj["slurmise"].record(cmd, job_name, slurm_id, step_id)


@main.command()
@click.option("--job-name", type=str, required=True, help="Name of the job")
@click.option("--slurm-id", type=str, required=True, help="SLURM id of job")
@click.option(
    "--numerical",
    type=str,
    help="Numerical run parameters in JSON format without outer {}, such as 'n:3,q:17.4'",
)
@click.option(
    "--categorical",
    type=str,
    help="Categorical run parameters in JSON format without outer {}",
)
@click.option("--cmd", type=str, help="Actual command run")
@click.pass_context
def raw_record(ctx, job_name, slurm_id, numerical, categorical, cmd):
    """Record a job"""
    # NOTE hack to get JSON parsing working with click who is too eager to try and process the args
    categorical = json.loads("{" + categorical + "}") if categorical else {}
    numerical = json.loads("{" + numerical + "}") if numerical else {}

    jd = job_data.JobData(
        job_name=job_name,
        slurm_id=slurm_id,
        numerical=numerical,
        categorical=categorical,
        cmd=cmd,
    )

    ctx.obj["slurmise"].raw_record(jd)


@main.command()
@click.pass_context
def print(ctx):
    ctx.obj["slurmise"].print()


@main.command()
@click.argument("cmd", nargs=1)
@click.option("--job-name", type=str, help="Name of the job")
@click.pass_context
def predict(ctx, cmd, job_name):
    query_jd, query_warns = ctx.obj["slurmise"].predict(cmd, job_name)
    click.echo(f"Predicted runtime: {query_jd.runtime}")
    click.echo(f"Predicted memory: {query_jd.memory}")
    if query_warns:
        click.echo(click.style("Warnings:", fg="yellow"), err=True, color="red")
        for warn in query_warns:
            click.echo(f"  {warn}", err=True)


@main.command()
@click.argument("cmd", nargs=1)
@click.option("--job-name", type=str, help="Name of the job")
@click.pass_context
def update_model(ctx, cmd, job_name):
    ctx.obj["slurmise"].update_model(cmd, job_name)


@main.command()
@click.pass_context
def update_all(ctx):
    ctx.obj["slurmise"].update_all_models()



import numpy as np
from click.testing import CliRunner

from slurmise import job_database
from slurmise.__main__ import main
from slurmise.job_data import JobData


def test_missing_toml():
    """Check that excluding a toml file will fail with error message."""
    runner = CliRunner()
    result = runner.invoke(
        main,
        ["record", "something"],
    )
    assert result.exit_code == 1
    assert "Slurmise requires a toml file" in result.output
    assert "See readme for more information" in result.output


def test_record(simple_toml, monkeypatch):
    mock_metadata = {
        "slurm_id": "1234",
        "step_id": "0",
        "job_name": "nupack",
        "state": "COMPLETED",
        "partition": "",
        "elapsed_seconds": 97201,
        "CPUs": 1,
        "memory_per_cpu": 0,
        "memory_per_node": 0,
        "max_rss": 232,
    }
    monkeypatch.setattr(
        "slurmise.slurm.parse_slurm_job_metadata", lambda *args, **kwargs: mock_metadata
    )

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--toml",
            simple_toml.toml,
            "record",
            "--slurm-id",
            "1234",
            "nupack monomer -T 2 -C simple",
        ],
    )
    assert result.exit_code == 0
    # test the job was successfully added
    with job_database.JobDatabase.get_database(simple_toml.db) as db:
        excepted_results = [
            JobData(
                job_name="nupack",
                slurm_id="1234.0",
                runtime=97201,
                memory=232,
                categorical={"complexity": "simple"},
                numerical={"threads": 2},
                cmd=None,
            ),
        ]

        query = JobData(
            job_name="nupack",
            categorical={"complexity": "simple"},
        )
        query_result = db.query(query)

        assert query_result == excepted_results


def test_raw_record(simple_toml):
    """Test the raw_record command."""
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--toml",
            simple_toml.toml,
            "raw-record",
            "--job-name",
            "test",
            "--slurm-id",
            "1234",
            "--numerical",
            '"n":3,"q":17.4',
            "--categorical",
            '"a":1,"b":2',
            "--cmd",
            "sleep 2",
        ],
    )
    assert result.exit_code == 0

    # test the job was successfully added
    with job_database.JobDatabase.get_database(simple_toml.db) as db:
        excepted_results = [
            JobData(
                job_name="test",
                slurm_id="1234",
                categorical={"a": 1, "b": 2},
                numerical={"n": 3, "q": 17.4},
                cmd=None,
            ),
        ]

        query = JobData(
            job_name="test",
            categorical={"a": 1, "b": 2},
        )
        query_result = db.query(query)

        assert query_result == excepted_results


def test_update_predict(nupack_toml):
    """Test the update and predict commands of slurmise.
    Initially, we run the update command to get the models for the nupack job.
    After the models are created, we run the predict command to predict the runtime and memory of a job.
    Two tests are run. The first predicts a runtime and memory values for a job that
    makes sense. The second test returns a runtime and memory values that are not
    possible. Because we cannot know the exact numbers we check of the expected strings.
    """
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--toml",
            nupack_toml.toml,
            "update-model",
            "nupack monomer -c 1 -S 4985",
        ],
        catch_exceptions=True,
    )
    if result.exception:
        print(f"Exception: {result.exception}")
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        [
            "--toml",
            nupack_toml.toml,
            "predict",
            "nupack monomer -c 3 -S 6543",
        ],
    )
    assert result.exit_code == 0
    tmp_stdout = result.stdout.split("\n")
    predicted_runtime = tmp_stdout[0].split(":")
    predicted_memory = tmp_stdout[1].split(":")
    assert "Predicted runtime" == predicted_runtime[0]
    np.testing.assert_allclose(float(predicted_runtime[1]), 9.29, rtol=0.01)
    assert "Predicted memory" == predicted_memory[0]
    np.testing.assert_allclose(float(predicted_memory[1]), 10168.72, rtol=0.01)

    # Test that slurmise returns the default values when the predicted values are not possible.
    result = runner.invoke(
        main,
        [
            "--toml",
            nupack_toml.toml,
            "predict",
            "nupack monomer -c 987654 -S 4985",
        ],
        catch_exceptions=True,
    )
    assert result.exit_code == 0
    tmp_stdout = result.stdout.split("\n")
    predicted_runtime = tmp_stdout[0].split(":")
    predicted_memory = tmp_stdout[1].split(":")
    assert "Predicted runtime" == predicted_runtime[0]
    assert float(predicted_runtime[1]) == 60
    assert "Predicted memory" == predicted_memory[0]
    assert float(predicted_memory[1]) == 1000
    assert "Warnings:" in result.stderr


def test_update_all_predict(nupack_toml):
    """Test the update all and predict commands of slurmise.
    Initially, we run the update command to get the models for the nupack job.
    After the models are created, we run the predict command to predict the runtime and memory of a job.
    Two tests are run. The first predicts a runtime and memory values for a job that
    makes sense. The second test returns a runtime and memory values that are not
    possible. Because we cannot know the exact numbers we check of the expected strings.
    """
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--toml",
            nupack_toml.toml,
            "update-all",
        ],
        catch_exceptions=True,
    )
    if result.exception:
        print(f"Exception: {result.exception}")
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        [
            "--toml",
            nupack_toml.toml,
            "predict",
            "nupack monomer -c 3 -S 6543",
        ],
    )
    assert result.exit_code == 0
    tmp_stdout = result.stdout.split("\n")
    predicted_runtime = tmp_stdout[0].split(":")
    predicted_memory = tmp_stdout[1].split(":")
    assert "Predicted runtime" == predicted_runtime[0]
    np.testing.assert_allclose(float(predicted_runtime[1]), 9.29, rtol=0.01)
    assert "Predicted memory" == predicted_memory[0]
    np.testing.assert_allclose(float(predicted_memory[1]), 10168.72, rtol=0.01)



import tomllib
from pathlib import Path
from collections import defaultdict
from slurmise import job_data
from slurmise.job_parse import file_parsers
from slurmise.job_parse.job_specification import JobSpec


class SlurmiseConfiguration:
    """SlurmiseConfiguration class parses and stores TOML configuration files for slurmise."""

    def __init__(self, toml_file: Path):
        """Parse a configuration TOML file"""
        self.file_parsers = {
            "file_size": file_parsers.FileSizeParser(),
            "file_lines": file_parsers.FileLinesParser(),
            "file_basename": file_parsers.FileBasename(),
            "file_md5": file_parsers.FileMD5(),
        }
        with open(toml_file, "rb") as f:
            toml_data = tomllib.load(f)

            self.slurmise_base_dir = toml_data["slurmise"]["base_dir"]
            Path(self.slurmise_base_dir).mkdir(parents=True, exist_ok=True)
            self.db_filename = Path(self.slurmise_base_dir) / toml_data["slurmise"].get(
                "db_filename", "slurmise.h5"
            )
            parsers = toml_data["slurmise"].get("file_parsers", {})

            for parser_name, config in parsers.items():
                return_type = config.get("return_type", "categorical")
                if "awk_script" in config:
                    script_is_file = config.get("script_is_file", False)
                    self.file_parsers[parser_name] = file_parsers.AwkParser(
                        parser_name,
                        return_type,
                        config["awk_script"],
                        script_is_file,
                    )

            self.jobs = toml_data["slurmise"].get("job", {})
            self.job_prefixes: dict[str, str] = {}
            self.default_runtime = defaultdict(lambda: int(toml_data["slurmise"].get("default_time", 60)))
            self.default_memory = defaultdict(lambda: int(toml_data["slurmise"].get("default_mem", 1000)))

            for job_name, job in self.jobs.items():
                self.jobs[job_name]["job_spec_obj"] = JobSpec(
                    job["job_spec"],
                    file_parsers=job.get("file_parsers", {}),
                    available_parsers=self.file_parsers,
                )
                if "job_prefix" in job:
                    self.job_prefixes[job_name] = job["job_prefix"]
                if "default_time" in job:
                    self.default_runtime[job_name] = int(job['default_time'])
                if "default_mem" in job:
                    self.default_memory[job_name] = int(job['default_mem'])

    def parse_job_cmd(
            self, cmd: str, job_name: str | None = None, slurm_id: str | None = None,
            step_id: str | None = None
            ) -> job_data.JobData:
        """Parse a job data dataset into a JobData object."""
        if job_name is None:  # try to infer
            for name, prefix in self.job_prefixes.items():
                if cmd.startswith(prefix):
                    job_name = name
                    cmd = cmd.removeprefix(prefix).lstrip()
                    break

            else:  # not a prefix. Runs when it does not hit the break.
                for name in self.jobs.keys():
                    if cmd.startswith(name):
                        job_name = name
                        cmd = cmd.removeprefix(name).lstrip()
                        break

                else:
                    raise ValueError(f"Unable to match job name to {cmd!r}")
        else:
            job_prefix = self.job_prefixes.get(job_name, None)
            if job_prefix is not None:
                cmd = cmd.removeprefix(job_prefix).lstrip()

        if job_name not in self.jobs:
            raise ValueError(f"Job {job_name} not found in configuration.")

        if step_id is not None:
            slurm_id = '.'.join([str(slurm_id), str(step_id)])
        jd = job_data.JobData(
            job_name=job_name,
            slurm_id=slurm_id,
            cmd=cmd
        )

        job_spec = self.jobs[job_name]["job_spec_obj"]
        return job_spec.parse_job_cmd(jd)

    def add_defaults(self, job_data: job_data.JobData) -> job_data.JobData:
        """Add default values to a job data object."""
        job_data.memory = self.default_memory[job_data.job_name]
        job_data.runtime = self.default_runtime[job_data.job_name]
        return job_data



from pathlib import Path
from slurmise import job_data
from slurmise.job_parse.file_parsers import FileParser
import re


# matches tokens like {threads:numeric}
JOB_SPEC_REGEX = re.compile(r"{(?:(?P<name>[^:}]+):)?(?P<kind>[^}]+)}")
KIND_TO_REGEX = {
    'file': '.+?',
    'gzip_file': '.+?',
    'file_list': '.+?',
    'numeric': '[-0-9.]+',
    'category': '.+?',
    'ignore': '.+?',
}
CATEGORICAL = "CATEGORICAL"
NUMERICAL = "NUMERICAL"


class JobSpec:
    def __init__(
            self,
            job_spec: str,
            file_parsers: dict[str, str] | None = None,
            available_parsers: dict[str, FileParser] | None = None,
        ):
        """Parse a job spec string into a regex with named capture groups.

        job_spec: The specification of parsing the supplied command.  Can contain
        placeholders for variables to parse as numerics, strings, or files.
        file_parsers: A dict of file variable names to parser names.  Can be a
        comma separate list or single string
        available_parsers: A dict of parser names to parser objects
        """
        self.job_spec_str = job_spec

        self.token_kinds = {}
        self.file_parsers: dict[str, list[FileParser]] = {}

        while match := JOB_SPEC_REGEX.search(job_spec):
            kind = match.group('kind')
            name = match.group('name')

            if kind not in KIND_TO_REGEX:
                raise ValueError(f"Token kind {kind} is unknown.")

            if kind == 'ignore':
                job_spec = job_spec.replace(match.group(0), f'{KIND_TO_REGEX[kind]}')

            else:
                if name is None:
                    raise ValueError(f"Token {match.group(0)} has no name.")
                self.token_kinds[name] = kind
                job_spec = job_spec.replace(match.group(0), f'(?P<{name}>{KIND_TO_REGEX[kind]})')

                if kind in ('file', 'gzip_file', 'file_list'):
                    self.file_parsers[name] = [
                        available_parsers[parser_type]
                        for parser_type in file_parsers[name].split(',')
                    ]

        self.job_regex = f'^{job_spec}$'

    def parse_job_cmd(self, job: job_data.JobData) -> job_data.JobData:
        m = re.match(self.job_regex, job.cmd)
        if m is None:
            raise ValueError(f"Job spec {self.job_spec_str} does not match command {job.cmd}.")

        for name, kind in self.token_kinds.items():
            if kind == 'numeric':
                job.numerical[name] = float(m.group(name))
            elif kind == 'category':
                job.categorical[name] = m.group(name)
            elif kind in ('file', 'gzip_file', 'file_list'):
                for parser in self.file_parsers[name]:
                    match kind:
                        case 'file':
                            file_value = parser.parse_file(Path(m.group(name)))
                        case 'gzip_file':
                            file_value = parser.parse_file(Path(m.group(name)), gzip_file=True)
                        case 'file_list':
                            file_value = [
                            parser.parse_file(Path(file.strip()))
                            for file in open(Path(m.group(name)), 'r')
                        ]

                    if parser.return_type == NUMERICAL:
                        job.numerical[f"{name}_{parser.name}"] = file_value
                    else:
                        job.categorical[f"{name}_{parser.name}"] = file_value

            else:
                raise ValueError(f"Unknown kind {kind}.")
            
        return job



import gzip
from dataclasses import dataclass, field
from pathlib import Path
import subprocess
import hashlib


NUMERICAL = "NUMERICAL"
CATEGORICAL = "CATEGORICAL"


@dataclass()
class FileParser:
    name: str = 'UNK'
    return_type: str = NUMERICAL

    def parse_file(self, path: Path, gzip_file: bool=False):
        raise NotImplementedError()


@dataclass()
class FileBasename(FileParser):
    def __init__(self):
        super().__init__(name='file_basename', return_type=CATEGORICAL)

    def parse_file(self, path: Path, gzip_file: bool=False):
        return path.name


@dataclass()
class FileMD5(FileParser):
    def __init__(self):
        super().__init__(name='file_md5', return_type=CATEGORICAL)

    def parse_file(self, path: Path, gzip_file: bool=False):
        md5_hash = hashlib.md5()
        md5_hash.update(path.read_bytes())
        # with path.read_bytes() as file:
        #     # Read the file in chunks to handle large files efficiently
        #     for chunk in iter(lambda: file.read(4096), b""):
        #         md5_hash.update(chunk)
        return md5_hash.hexdigest()


@dataclass()
class FileSizeParser(FileParser):
    def __init__(self):
        super().__init__(name='file_size', return_type=NUMERICAL)

    def parse_file(self, path: Path, gzip_file: bool=False):
        return path.stat().st_size  # in bytes


@dataclass()
class FileLinesParser(FileParser):
    def __init__(self):
        super().__init__(name='file_lines', return_type=NUMERICAL)

    def parse_file(self, path: Path, gzip_file: bool=False):
        if gzip_file:
            with gzip.open(path, 'rb') as infile:
                # gzip files have no raw read, use slower loop
                for lines, _ in enumerate(infile):
                    pass
                return lines + 1
        else:
            with open(path, 'rb') as infile:
                lines = 1  # will count the last line as well.  Off by one for empty files
                buf_size = 1024 * 1024
                read_f = infile.raw.read

                while buf := read_f(buf_size):
                    lines += buf.count(b'\n')

            return lines


@dataclass()
class AwkParser(FileParser):
    args: list[str] = field(default_factory=list)

    def __init__(self, name, return_type, script, script_is_file=False):
        return_type = return_type.upper()
        super().__init__(name=name, return_type=return_type)
        self.args = ['awk', script]
        if script_is_file:
            # add file argument to awk
            self.args.insert(1, '-f')

    def parse_file(self, path: Path, gzip_file: bool=False):
        if gzip_file:
            zcat = subprocess.Popen(('zcat', path), stdout=subprocess.PIPE)
            result = subprocess.check_output(self.args, stdin=zcat.stdout, text=True)
            zcat.wait()
        else:
            result = subprocess.check_output(self.args + [path], text=True)

        if self.return_type == NUMERICAL:
            return [float(token) for token in result.split()]
        return result.strip()



import numpy as np
import pytest

from slurmise.job_data import JobData
from slurmise.job_database import JobDatabase


@pytest.fixture
def empty_h5py_file(tmp_path):
    d = tmp_path
    p = d / "slurmise.h5"
    return p


@pytest.fixture
def small_db(empty_h5py_file):
    with JobDatabase.get_database(empty_h5py_file) as db:
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="1",
                runtime=5,
                memory=100,
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="2",
                runtime=6,
                memory=128,
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="1",
                runtime=5,
                memory=100,
                categorical={"option1": "value1", "option2": "value2"},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="2",
                numerical={"filesizes": [123, 512, 128]},
                categorical={"option1": "value2"},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="3",
            )
        )
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="4",
                runtime=7,
                memory=100,
                categorical={"option2": "value2", "option1": "value1"},
            )
        )
        yield db


def test_close(empty_h5py_file):
    """Test opening and closing an empty database."""
    with JobDatabase.get_database(empty_h5py_file):
        pass


def test_rqd_flat(empty_h5py_file):
    """Test record, query and delete on jobs without categorical values."""
    with JobDatabase.get_database(empty_h5py_file) as db:
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="1",
                runtime=5,
                memory=100,
            )
        )
        # assert commit_value == 1

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="2",
                runtime=6,
                memory=128,
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        excepted_results = [
            JobData(job_name="test_job", slurm_id="1", runtime=5, memory=100),
            JobData(
                job_name="test_job",
                slurm_id="2",
                runtime=6,
                memory=128,
                numerical={"filesizes": np.array([123, 512, 128])},
            ),
        ]

        query_result = db.query(JobData(job_name="test_job"))

        assert query_result == excepted_results

        db.record(
            JobData(
                job_name="test_job2",
                slurm_id="2",
                runtime=6,
                memory=128,
                numerical={"filesizes": [123, 512, 128]},
            )
        )
        db.delete(JobData(job_name="test_job"))
        query_result = db.query(JobData(job_name="test_job"))
        assert query_result == []
        query_result = db.query(JobData(job_name="test_job2"))
        excepted_results = [
            JobData(
                slurm_id="2",
                job_name="test_job2",
                runtime=6,
                memory=128,
                numerical={"filesizes": np.array([123, 512, 128])},
            )
        ]
        np.testing.assert_equal(query_result, excepted_results)


def test_rqd_with_categories(empty_h5py_file):
    with JobDatabase.get_database(empty_h5py_file) as db:
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="1",
                runtime=5,
                memory=100,
                categorical={"option1": "value1", "option2": "value2"},
            )
        )
        # assert commit_value == 1

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="2",
                numerical={"filesizes": [123, 512, 128]},
                categorical={"option1": "value2"},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="3",
            )
        )
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="4",
                runtime=7,
                memory=100,
                categorical={"option2": "value2", "option1": "value1"},
            )
        )

        excepted_results = [
            JobData(job_name="test_job", slurm_id="3"),
        ]

        query_result = db.query(JobData(job_name="test_job"))
        assert query_result == excepted_results

        excepted_results = [
            JobData(
                job_name="test_job",
                slurm_id="1",
                runtime=5,
                memory=100,
                categorical={"option1": "value1", "option2": "value2"},
            ),
            JobData(
                job_name="test_job",
                slurm_id="4",
                runtime=7,
                memory=100,
                categorical={"option2": "value2", "option1": "value1"},
            ),
        ]

        query_result = db.query(
            JobData(
                job_name="test_job",
                categorical={"option2": "value2", "option1": "value1"},
            )
        )
        assert query_result == excepted_results

        query_result = db.query(
            JobData(
                job_name="test_job",
                categorical={
                    "option1": "value1",
                    "option2": "value2",
                },
            )
        )
        assert query_result == excepted_results

        excepted_results = [
            JobData(
                job_name="test_job",
                slurm_id="2",
                numerical={"filesizes": np.array([123, 512, 128])},
                categorical={"option1": "value2"},
            )
        ]
        query_result = db.query(
            JobData(
                job_name="test_job",
                categorical={
                    "option1": "value2",
                },
            )
        )
        assert query_result == excepted_results


def test_rqd_with_emptydb(empty_h5py_file):
    """Test record, query and delete on empty database."""
    with JobDatabase.get_database(empty_h5py_file) as db:
        query_result = db.query(
            JobData(
                job_name="test_job",
                categorical={
                    "option1": "value2",
                },
            )
        )
        assert query_result == []

        query_result = db.query(
            JobData(
                job_name="test_job2",
            )
        )
        assert query_result == []

        # no errors
        db.delete(JobData(job_name="test_job3"), delete_all_children=False)
        db.delete(
            JobData(job_name="test_job3", categorical={"option1": "value1"}),
            delete_all_children=False,
        )


def test_iterate_database(small_db):
    """Test access to all jobs and data through an iterator."""
    expected_queries = [
        JobData(job_name='test_job', categorical={'option1': 'value1', 'option2': 'value2'}),
        JobData(job_name='test_job', categorical={'option1': 'value1'}),
        JobData(job_name='test_job', categorical={'option1': 'value2'}),
        JobData(job_name='test_job'),
    ]

    for (query, jobs), expected in zip(small_db.iterate_database(), expected_queries, strict=True):
        assert query == expected
        expected_result = small_db.query(query)
        assert jobs == expected_result


def test_delete(small_db):
    """Test deletion of jobs with categorical where delete_all_children is False."""
    expected_result = [
        JobData(
            job_name="test_job",
            slurm_id="1",
            memory=np.int64(100),
            runtime=np.int64(5),
        ),
        JobData(
            job_name="test_job",
            slurm_id="2",
            numerical={"filesizes": np.array([123, 512, 128])},
            memory=np.int64(128),
            runtime=np.int64(6),
        ),
        JobData(
            job_name="test_job",
            slurm_id="3",
        ),
    ]

    simple_query = JobData(job_name="test_job")
    query_result = small_db.query(simple_query)
    assert query_result == expected_result

    small_db.delete(simple_query, delete_all_children=False)

    query_result = small_db.query(simple_query)
    assert query_result == []

    excepted_results = [
        JobData(
            job_name="test_job",
            slurm_id="1",
            runtime=5,
            memory=100,
            categorical={"option1": "value1", "option2": "value2"},
        ),
        JobData(
            job_name="test_job",
            slurm_id="4",
            runtime=7,
            memory=100,
            categorical={"option2": "value2", "option1": "value1"},
        ),
    ]

    query = JobData(
        job_name="test_job",
        categorical={"option1": "value1", "option2": "value2"},
    )
    query_result = small_db.query(query)

    assert query_result == excepted_results

    # Missing category values results to noop delete
    small_db.delete(
        JobData(job_name="test_job", categorical={"option1": "value1"}),
        delete_all_children=False,
    )

    query_result = small_db.query(query)

    assert query_result == excepted_results

    small_db.delete(query, delete_all_children=False)

    query_result = small_db.query(query)

    assert query_result == []


def test_delete_all_children(small_db):
    query_result = small_db.query(
        JobData(
            job_name="test_job", categorical={"option1": "value1", "option2": "value2"}
        )
    )
    assert len(query_result)

    small_db.delete(
        JobData(job_name="test_job", categorical={"option1": "value1"}),
        delete_all_children=True,
    )

    query_result = small_db.query(
        JobData(
            job_name="test_job", categorical={"option1": "value1", "option2": "value2"}
        )
    )
    assert query_result == []

    query_result = small_db.query(JobData(job_name="test_job"))
    assert len(query_result)

    query_result = small_db.query(
        JobData(job_name="test_job", categorical={"option1": "value2"})
    )
    assert len(query_result)
    small_db.delete(JobData(job_name="test_job"), delete_all_children=True)

    query_result = small_db.query(JobData(job_name="test_job"))
    assert query_result == []

    query_result = small_db.query(
        JobData(job_name="test_job", categorical={"option1": "value2"})
    )
    assert query_result == []


def test_update_missing_mem_elapsed(empty_h5py_file, monkeypatch):
    def mock_parse_slurm_job_metadata(slurm_id, step_id):
        return {
            "max_rss": 101,
            "elapsed_seconds": 100,
        }

    monkeypatch.setattr(
        "slurmise.job_database.slurm.parse_slurm_job_metadata",
        mock_parse_slurm_job_metadata,
    )

    with JobDatabase.get_database(empty_h5py_file) as db:
        db.record(
            JobData(
                job_name="test_job",
                slurm_id="1.0",
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="2.1",
                memory=128,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="3.extern",
                runtime=111,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        db.record(
            JobData(
                job_name="test_job",
                slurm_id="4.extern",
                memory=138,
                runtime=222,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": [123, 512, 128]},
            )
        )

        expected_output = [
            JobData(
                job_name="test_job",
                slurm_id="1.0",
                memory=101,
                runtime=100,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": np.array([123, 512, 128])},
            ),
            JobData(
                job_name="test_job",
                slurm_id="2.1",
                memory=128,
                runtime=100,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": np.array([123, 512, 128])},
            ),
            JobData(
                job_name="test_job",
                slurm_id="3.extern",
                runtime=111,
                memory=101,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": np.array([123, 512, 128])},
            ),
            JobData(
                job_name="test_job",
                slurm_id="4.extern",
                memory=138,
                runtime=222,
                categorical={"option1": "value1", "option2": "value2"},
                numerical={"filesizes": np.array([123, 512, 128])},
            ),
        ]

        results = db.query(
            JobData(
                job_name="test_job",
                categorical={"option1": "value1", "option2": "value2"},
            ),
            update_missing=True,
        )
        assert results == expected_output

        results = db.query(
            JobData(
                job_name="test_job",
                categorical={"option1": "value1", "option2": "value2"},
            ),
            update_missing=False,
        )
        assert results == expected_output



from dataclasses import asdict

import numpy as np
import pandas as pd

from .job_data import JobData


def jobs_to_pandas(jobs: list[JobData]):
    """
    Convert a list of JobData objects to a pandas DataFrame. The DataFrame will have
    columns for each categorical and numerical feature, and will not include the job_name,
    slurm_id, memory, or runtime columns.

    :param jobs: A list of JobData objects
    :type jobs: list[JobData]
    :return: A pandas DataFrame with columns for each categorical and numerical feature
    :rtype: pd.DataFrame

    """

    df = pd.json_normalize([asdict(job) for job in jobs])

    # Convert categorical columns to category type
    for col in df.columns:
        if col.startswith("categorical."):
            df[col] = df[col].astype("category")

    # Rename the categorical columns, drop .categorical prefix
    df.columns = [col.replace("categorical.", "") for col in df.columns]

    # Do the same for numerical columns
    for col in df.columns:
        new_col_name = col.replace("numerical.", "")

        if col.startswith("numerical."):
            if df[col].dtype == "object":
                # Check if they are all numpy arrays
                if all([isinstance(row, np.ndarray) for row in df[col]]):
                    # Check if the column is a numpy array of all the same size
                    sizes = set([row.shape for row in df[col]])

                    if len(sizes) == 1:
                        # If all the same size, expand each element of the numpy array into a new column
                        col_df = pd.DataFrame(
                            np.vstack(
                                [
                                    s.flatten()
                                    for s in df.loc[0:10, "numerical.sequences"]
                                ]
                            )
                        )
                        col_df.columns = [
                            f"{new_col_name}_{i}" for i in range(col_df.shape[1])
                        ]

                        # Drop the original column and add the new columns
                        df = df.drop(columns=[col])
                        df = pd.concat([df, col_df], axis=1)

                    else:
                        raise ValueError(
                            f"Numerical feature {new_col_name} is an a numpy array of different sizes. "
                            f"Numpy arrays are supported only if they are all the same size."
                        )

                else:
                    raise ValueError(
                        "Numerical columns must be scalars or equal length numpy arrays"
                    )

    df.columns = [col.replace("numerical.", "") for col in df.columns]

    # Get the numerical columns
    df = df.drop(columns=["job_name", "slurm_id", "cmd"])

    # Transform features
    categorical_features = [name for name in df.columns if df[name].dtype == "category"]
    numerical_features = [
        name
        for name in df.columns
        if name not in categorical_features and name not in ["memory", "runtime"]
    ]

    return df, categorical_features, numerical_features



import contextlib
import dataclasses
import os
from typing import Any
import time

import h5py
import numpy as np

from slurmise import slurm
from slurmise.job_data import JobData


class JobDatabase:
    """
    This class creates the database to store job information.
    It saves the database in HDF5 file.
    """

    def __init__(self, db_file: str, max_retries: int = 5):
        """
        The DB file is and HDF5 file.
        Use **get_database** and a context manager to have the file automatically
        closed.
        """

        attempt = 0
        while True:
            attempt += 1
            try:
                self.db = h5py.File(db_file, "a")
                break
            except BlockingIOError:
                if attempt > max_retries:
                    raise
                time.sleep(attempt)

    def _close(self):
        self.db.close()

    @staticmethod
    @contextlib.contextmanager
    def get_database(db_file: str, max_retries: int = 5) -> "JobDatabase":
        """
        Use in context manager to automatically open and close db file.

        :arguments:

            :db_file: HDF5 file to use as database

        :yields:

            JobDatabase with opened db file

        :finally:

            Closes h5py database
        """

        db = JobDatabase(db_file, max_retries)
        try:
            yield db
        finally:
            db._close()

    def record(self, job_data: JobData, ignore_existing_job: bool = False) -> None:
        """
        It records JobData information in the database. A tree is created based
        on the job name, categorical values and slurm id. The leaves of the tree
        are the memory, runtime and numericals of the JobData.
        """
        table_name = JobDatabase.get_table_name(job_data)
        if ignore_existing_job and self.job_exists(job_data):
            return

        table = self.db.require_group(name=table_name)

        if job_data.memory is not None:
            val = np.asarray(job_data.memory)
            _ = table.create_dataset(name="memory", shape=val.shape, data=val)

        if job_data.runtime is not None:
            val = np.asarray(job_data.runtime)
            _ = table.create_dataset(name="runtime", shape=val.shape, data=val)

        for var, value in job_data.numerical.items():
            val = np.asarray(value)
            _ = table.create_dataset(name=var, shape=val.shape, data=val)

    def job_exists(self, job_data: JobData) -> bool:
        table_name = JobDatabase.get_table_name(job_data)
        return table_name in self.db

    def update(self, **kargs):
        raise NotImplementedError("Later feature")

    def query(self, job_data: JobData, update_missing: bool = False) -> list[JobData]:
        """
        Query returns a list of JobData objects based on the requested JobData.
        The returned jobs match the query JobData's job name and categoricals.
        `update_missing` will try to get maxRSS and elapsed from sacct if not found in the DB.

        Note: It does not decent into all child categories, only the highest matching leaves
        """
        group_name = JobDatabase.get_group_name(job_data)

        job_group = self.db.get(group_name, default={})
        result = []
        for slurm_id, slurm_data in job_group.items():
            if JobDatabase.is_slurm_job(slurm_data):
                result.append(
                    JobData.from_dataset(
                        job_name=job_data.job_name,
                        slurm_id=slurm_id,
                        categorical=job_data.categorical,
                        dataset=slurm_data,
                    )
                )

        if update_missing:
            result = self.update_missing_data(result)

        return result

    def delete(self, job_data: JobData, delete_all_children: bool = False) -> None:
        """
        Delete jobs with matching job name and categoricals.

        :arguments:

            :job_data: JobData object with name and categorical which should be removed.
            :delete_all_children: When true, will delete recursively any matching jobs
        """
        group_name = JobDatabase.get_group_name(job_data)

        if group_name in self.db:
            if delete_all_children:
                del self.db[group_name]
            else:
                # Traverse and delete only Datasets
                job_group = self.db.get(group_name, default={})
                for slurm_id, slurm_data in job_group.items():
                    if JobDatabase.is_slurm_job(slurm_data):
                        del job_group[slurm_id]

    def clear(self):
        raise NotImplementedError("Empting the DB is not yet supported")

    def record_fit(self, fit):
        raise NotImplementedError("Storing fits is not supported yet")

    def query_fit(self, fit):
        raise NotImplementedError("Storing fits is not supported yet")

    def update_missing_data(self, jobs: list[JobData]) -> list[JobData]:
        """
        Update missing mem and runtime for jobs with incomplete data in the db.
        Takes a list of JobData which was queried from the db, updates the db, and returns the updated job list.

        TODO: gather slurm_ids of jobs that need updating and do it in one call
        """
        updated_jobs = []
        for job in jobs:
            if job.memory is None or job.runtime is None:
                if "." in job.slurm_id:
                    slurm_id, step_id = job.slurm_id.split(".")
                else:
                    slurm_id, step_id = job.slurm_id, None
                job_info = slurm.parse_slurm_job_metadata(slurm_id = slurm_id, step_id = step_id)

                # job dataclass is immutable, so this creates a new object with the updated values
                # ternary's are to avoid updating if the value is already present which causes a "dataset already exists" error
                self.record(
                    dataclasses.replace(
                        job,
                        memory=job_info["max_rss"] if job.memory is None else None,
                        runtime=(
                            job_info["elapsed_seconds"] if job.runtime is None else None
                        ),
                        numerical={},
                    )
                )

                job = dataclasses.replace(
                    job,
                    memory=job_info["max_rss"] if job.memory is None else job.memory,
                    runtime=(
                        job_info["elapsed_seconds"]
                        if job.runtime is None
                        else job.runtime
                    ),
                )

            updated_jobs.append(job)

        return updated_jobs

    @staticmethod
    def get_table_name(job_data: JobData) -> str:
        table_name = JobDatabase.get_group_name(job_data)
        table_name += f"/{job_data.slurm_id}"
        return table_name

    @staticmethod
    def get_group_name(job_data: JobData) -> str:
        group_name = f"/{job_data.job_name}"
        for key in sorted(job_data.categorical.keys()):
            group_name += f"/{key}={job_data.categorical[key]}"
        return group_name

    @staticmethod
    def is_dataset(f: Any) -> bool:
        """
        Test if object is an h5py Dataset
        """
        return type(f) == h5py._hl.dataset.Dataset

    @staticmethod
    def is_slurm_job(f: Any) -> bool:
        """
        Test if object is non-empty or its first element is a Dataset.
        This is consistent with a slurm job
        """

        if len(f) == 0:  # group contains no values
            return True
        first_element = f[list(f.keys())[0]]
        # group contains a dataset
        return JobDatabase.is_dataset(first_element)

    def print(self):
        JobDatabase.print_hdf5(self.db)

    def iterate_database(self, update_missing: bool = False):
        """
        Yield key (query job) value (list of jobs) pairs of entire database.
        """
        for job_name in self.db.keys():
            entry = self.db[job_name]
            for categoricals, jobs in JobDatabase.iterate_jobs(entry):
                categoricals = dict(cat.split("=") for cat in categoricals)
                query = JobData(job_name=job_name, categorical=categoricals)
                jobs = [
                    JobData.from_dataset(
                        job_name=query.job_name,
                        slurm_id=slurm_id,
                        categorical=query.categorical,
                        dataset=slurm_data,
                    )
                    for slurm_id, slurm_data in jobs.items()
                ]
                if update_missing:
                    jobs = self.update_missing_data(jobs)

                yield query, jobs

    @staticmethod
    def iterate_jobs(h5py_obj, categoricals=None):
        if categoricals is None:
            categoricals = tuple()

        jobs = {}
        for key, entry in h5py_obj.items():
            if JobDatabase.is_slurm_job(entry):
                jobs[key] = entry
            else:
                yield from JobDatabase.iterate_jobs(entry, categoricals + (key,))
        yield categoricals, jobs

    @staticmethod
    def print_hdf5(
        h5py_obj, level=-1, print_full_name: bool = False, print_attrs: bool = True
    ) -> None:
        """Prints the name and shape of datasets in a H5py HDF5 file.

        Parameters
        ----------
        h5py_obj: [h5py.File, h5py.Group]
            the h5py.File or h5py.Group object
        level: int
            What level of the file tree you are in
        print_full_name
            If True, the full tree will be printed as the name, e.g. /group0/group1/group2/dataset: ...
            If False, only the current node will be printed, e.g. dataset:
        print_attrs
            If True: print all attributes in the file
        Returns
        -------
        None

        """

        def is_group(f):
            return type(f) == h5py._hl.group.Group

        def is_dataset(f):
            return type(f) == h5py._hl.dataset.Dataset

        def print_level(level, n_spaces=5) -> str:
            if level == -1:
                return ""
            prepend = "|" + " " * (n_spaces - 1)
            prepend *= level
            tree = "|" + "-" * (n_spaces - 2) + " "
            return prepend + tree

        for key in h5py_obj.keys():
            entry = h5py_obj[key]
            name = entry.name if print_full_name else os.path.basename(entry.name)
            if is_group(entry):
                print(f"{print_level(level)}{name}")
                JobDatabase.print_hdf5(
                    entry, level + 1, print_full_name=print_full_name
                )
            elif is_dataset(entry):
                shape = entry.shape
                dtype = entry.dtype
                print(f"{print_level(level)}{name}: {shape} {dtype} {entry[()]}")
        if level == -1:
            if print_attrs:
                print("attrs: ")
                for key, value in h5py_obj.attrs.items():
                    print(f" {key}: {value}")



import pytest
import gzip
from slurmise.job_parse.job_specification import JobSpec
from slurmise.job_parse import file_parsers
from slurmise.job_data import JobData


# Tests for JobSpec
def test_job_spec_named_ignore():
    spec = JobSpec('cmd -T {threads:numeric} -i {named:ignore}')
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd='cmd -T 10 -i asdf',
            ))
    assert jd == JobData(
        job_name='test',
        cmd='cmd -T 10 -i asdf',
        numerical={'threads': 10},
    )


def test_job_spec_unknown_kind():
    with pytest.raises(ValueError, match="Token kind double is unknown"):
        JobSpec('cmd -T {threads:double}')

def test_job_spec_token_with_no_name():
    with pytest.raises(ValueError, match="Token {numeric} has no name."):
        JobSpec('cmd -T {numeric}')

def test_job_spec_with_builtin_parsers_basename(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "file_basename"
    '''
    available_parsers = {
        'file_basename': file_parsers.FileBasename(),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'file_basename'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {}
    assert jd.categorical == {'input1_file_basename': 'input.txt'}


def test_job_spec_with_builtin_parsers_md5hash(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "file_md5"
    '''
    available_parsers = {
        'file_md5': file_parsers.FileMD5(),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'file_md5'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
        '''here is
        some lines
        of text'''
    )
    test_file = tmp_path / 'test.txt'
    test_file.write_text(
        '''here is
        some lines
        of text'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {}

    jd_test = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=f"--input1 {test_file}",
            ))
    # test that md5 digest reflects file content
    assert jd.categorical == jd_test.categorical


def test_job_spec_with_builtin_parsers(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file} --input2 {input2:file}"
        file_parsers.input1 = "file_lines"
        file_parsers.input2 = "file_size"
    '''

    available_parsers = {
        'file_lines': file_parsers.FileLinesParser(),
        'file_size': file_parsers.FileSizeParser(),
    }

    spec = JobSpec(
        "--input1 {lines:file} --input2 {filesize:file}",
        file_parsers={'lines': 'file_lines', 'filesize': 'file_size'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
        '''here is
        some lines
        of text'''
    )

    command = f"--input1 {input_file} --input2 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {'lines_file_lines': 3, 'filesize_file_size': 42}

def test_job_spec_with_builtin_parsers_gzipped(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:gzip_file} --input2 {input2:gzip_file}"
        file_parsers.input1 = "file_lines"
        file_parsers.input2 = "file_size"
    '''

    available_parsers = {
        'file_lines': file_parsers.FileLinesParser(),
        'file_size': file_parsers.FileSizeParser(),
    }

    spec = JobSpec(
        "--input1 {lines:gzip_file} --input2 {filesize:gzip_file}",
        file_parsers={'lines': 'file_lines', 'filesize': 'file_size'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt.gz'
    with gzip.open(input_file, 'wt') as infile:
        for _ in range(100):
            infile.write(
                '''here is
                some lines
                of text'''
            )

    command = f"--input1 {input_file} --input2 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {'lines_file_lines': 201, 'filesize_file_size': 99}

def test_job_spec_with_builtin_parsers_file_list(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file_list}"
        file_parsers.input1 = "file_lines,file_size"
    '''

    available_parsers = {
        'file_lines': file_parsers.FileLinesParser(),
        'file_size': file_parsers.FileSizeParser(),
    }

    spec = JobSpec(
        "--input1 {lines:file_list}",
        file_parsers={'lines': 'file_lines,file_size'},
        available_parsers=available_parsers,
    )

    file_list = tmp_path / 'listing.txt'
    with file_list.open('w') as fl:
        for i in range(5):
            input_file = tmp_path / f'input_{i}.txt'
            fl.write(f'{input_file}\n')
            input_file.write_text(
                '''here is
                some lines
                of text''' * (5*(i+1))
            )

    command = f"--input1 {file_list}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {
        'lines_file_lines': [10*i + 1 for i in range(1, 6)],
        'lines_file_size': [290*i for i in range(1, 6)],
    }

def test_job_spec_with_multiple_builtin_parsers(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "file_lines,file_size"
    '''

    available_parsers = {
        'file_lines': file_parsers.FileLinesParser(),
        'file_size': file_parsers.FileSizeParser(),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'file_lines,file_size'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
        '''here is
        some lines
        of text'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.numerical == {'input1_file_lines': 3, 'input1_file_size': 42}

def test_job_spec_with_awk_parsers(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:gzip_file}"
        job_spec = "--input1 {input1:file_list}"
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "epochs,network"

        [slurmise.file_parsers.epochs]
        return_type = "numerical"
        awk_script = "/^epochs:/ {print $2}"

        [slurmise.file_parsers.network]
        return_type = "categorical"
        awk_script = "/^network type:/ {print $3}"
    '''

    available_parsers = {
        'epochs': file_parsers.AwkParser('epochs', 'numerical', '/^epochs:/ {print $2 ; exit}'),
        'network': file_parsers.AwkParser('network', 'categorical', '/^network type:/ {print $3 ; exit}'),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'epochs,network'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
'''epochs: 12
network type: conv_NN
network type: IGNORED!
some more text'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.categorical == {'input1_network': 'conv_NN'}
    assert jd.numerical == {'input1_epochs': [12]}

def test_job_spec_with_awk_parsers_multiple_numerics(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "layers"

        [slurmise.file_parsers.layers]
        return_type = "numerical"
        awk_script = "/^layers:/ {print $2}"
    '''

    available_parsers = {
        'layers': file_parsers.AwkParser('layers', 'numerical', '/^layers:/ {$1="" ; print $0}'),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'layers'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
'''layers: 12
layers: 14
layers: 16
layers: 18 24 36
some more text'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.categorical == {}
    assert jd.numerical == {'input1_layers': [12, 14, 16, 18, 24, 36]}


def test_job_spec_with_awk_file(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:file}"
        file_parsers.input1 = "fasta_inline,fasta_script"

        [slurmise.file_parsers.fasta_inline]
        return_type = "numerical"
        awk_script = "/^layers:/ {print $2}"

        [slurmise.file_parsers.fasta_script]
        return_type = "numerical"
        awk_script = "/path/to/awk/file.awk"
        script_is_file = True
    '''

    awk_script = ''' /^>/ {if (seq) print seq; seq=0} 
/^>/ {next} 
{seq = seq + length($0)} 
END {if (seq) print seq}
'''
    awk_file = tmp_path / 'parse_fasta.awk'
    awk_file.write_text(awk_script)

    available_parsers = {
        'fasta_inline': file_parsers.AwkParser(
            'fasta_inline', 'numerical', awk_script),
        'fasta_script': file_parsers.AwkParser(
            'fasta_script', 'numerical', awk_file, script_is_file=True),
    }

    spec = JobSpec(
        "--input1 {input1:file}",
        file_parsers={'input1': 'fasta_inline,fasta_script'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt'
    input_file.write_text(
'''>sequence 1
1234567890
1234567890
1234567890
1234567890
>sequence 2
1234567890
1234567890
12345
>sequence 3
1234567890
1234567890
1234567890
1234567890
123
>sequence 4
1
'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.categorical == {}
    assert jd.numerical == {
        'input1_fasta_inline': [40, 25, 43, 1],
        'input1_fasta_script': [40, 25, 43, 1],
    }


def test_job_spec_with_awk_gzip_file(tmp_path):
    '''
        [slurmise.job.builtin_files]
        job_spec = "--input1 {input1:gzip_file}"
        file_parsers.input1 = "fasta_inline,fasta_script"

        [slurmise.file_parsers.fasta_inline]
        return_type = "numerical"
        awk_script = "/^layers:/ {print $2}"

        [slurmise.file_parsers.fasta_script]
        return_type = "numerical"
        awk_script = "/path/to/awk/file.awk"
        script_is_file = True
    '''

    awk_script = ''' /^>/ {if (seq) print seq; seq=0} 
/^>/ {next} 
{seq = seq + length($0)} 
END {if (seq) print seq}
'''
    awk_file = tmp_path / 'parse_fasta.awk'
    awk_file.write_text(awk_script)

    available_parsers = {
        'fasta_inline': file_parsers.AwkParser(
            'fasta_inline', 'numerical', awk_script),
        'fasta_script': file_parsers.AwkParser(
            'fasta_script', 'numerical', awk_file, script_is_file=True),
    }

    spec = JobSpec(
        "--input1 {input1:gzip_file}",
        file_parsers={'input1': 'fasta_inline,fasta_script'},
        available_parsers=available_parsers,
    )

    input_file = tmp_path / 'input.txt.gz'
    with gzip.open(input_file, 'wt') as infile:
        infile.write(
'''>sequence 1
1234567890
1234567890
1234567890
1234567890
>sequence 2
1234567890
1234567890
12345
>sequence 3
1234567890
1234567890
1234567890
1234567890
123
>sequence 4
1
'''
    )

    command = f"--input1 {input_file}"
    jd = spec.parse_job_cmd(
        JobData(
            job_name='test',
            cmd=command,
            ))
    assert jd.job_name == 'test'
    assert jd.categorical == {}
    assert jd.numerical == {
        'input1_fasta_inline': [40, 25, 43, 1],
        'input1_fasta_script': [40, 25, 43, 1],
    }
